// SetClip.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

void SetClipFromFile(const TCHAR *szFilename, const bool bSingleLine, const bool bAppend);
void EmptyClip();
void SetClipFromArguments(const int startArg, const int argc, _TCHAR* argv[], const bool bSingleLine, const bool bAppend);
bool GetClipboardText(std::basic_string<TCHAR> &strClipboardText);

int _tmain(int argc, _TCHAR* argv[])
{
	bool bShowUsage = false;

	if (argc <= 1)
	{
		bShowUsage = true;
	}
	else if (0 == _tcsicmp(_T("-f"), argv[1]))
	{
		bool bSingleLine = false;
		bool bAppend = false;
		int startArg = 2;

		while(startArg < argc)
		{
			if (0 == _tcsicmp(_T("--"), argv[startArg]))
			{
				++startArg;
				break;
			}
			else if (0 == _tcsicmp(_T("-l"), argv[startArg]))
			{
				++startArg;
				bSingleLine = true;
			}
			else if (0 == _tcsicmp(_T("-a"), argv[startArg]))
			{
				++startArg;
				bAppend = true;
			}
			else
			{
				break;
			}
		}

		if (startArg == (argc-1))
		{
			SetClipFromFile(argv[startArg], bSingleLine, bAppend);
		}
		else
		{
			bShowUsage = true;
		}
	}
	else if (0 == _tcsicmp(_T("-e"), argv[1]))
	{
		if (argc == 2)
		{
			EmptyClip();
		}
		else
		{
			bShowUsage = true;
		}
	}
	else
	{
		bool bSingleLine = false;
		bool bAppend = false;
		int startArg = 1;

		while(startArg < argc)
		{
			if (0 == _tcsicmp(_T("--"), argv[startArg]))
			{
				++startArg;
				break;
			}
			else if (0 == _tcsicmp(_T("-l"), argv[startArg]))
			{
				++startArg;
				bSingleLine = true;
			}
			else if (0 == _tcsicmp(_T("-a"), argv[startArg]))
			{
				++startArg;
				bAppend = true;
			}
			else
			{
				break;
			}
		}

		if (startArg < argc)
		{
			SetClipFromArguments(startArg, argc, argv, bSingleLine, bAppend);
		}
	}

	if (bShowUsage)
	{
		_tprintf(_T("\nUsage 1: SetClip [-l] [-a] [--] <arg1> <arg2> ...\n\n"));
		_tprintf(_T("The arguments will be put in the clipboard with each one on a new line.\n"));
		_tprintf(_T("Use quotes around arguments containing spaces that you want on the same line.\n"));
		_tprintf(_T("Use \"\" for a blank line.\n"));
		_tprintf(_T("Give the -l argument to output everything on one line instead.\n"));
		_tprintf(_T("Give the -a argument to append to any existing clipboard text.\n"));
		_tprintf(_T("Give the -- argument if you want to actually put \"-l\", \"-f\", \"-e\" or \"-a\" into the clipboard.\n\n"));
		_tprintf(_T("Usage 2: SetClip -f [-l] [-a] [--] <filename>\n\n"));
		_tprintf(_T("The file will be read and put into the clipboard as text.\n"));
		_tprintf(_T("Give the -l argument to avoid adding a return when appending to existing text.\n"));
		_tprintf(_T("Give the -a argument to append to any existing clipboard text.\n\n"));
		_tprintf(_T("Give the -- argument if you want to actually put a file called \"-l\", \"-f\", \"-e\" or \"-a\" into the clipboard.\n\n"));
		_tprintf(_T("Usage 3: SetClip -e\n\n"));
		_tprintf(_T("The clipboard will be emptied.\n"));
	}

	return 0;
}

void SetClipFromFile(const TCHAR *szFilename, const bool bSingleLine, const bool bAppend)
{
	std::basic_string<TCHAR> strClipboardText;

	if (bAppend)
	{
		if (!GetClipboardText(strClipboardText))
		{
			strClipboardText.clear();
		}
	}

	HANDLE hFile = CreateFile(szFilename, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

	if (INVALID_HANDLE_VALUE == hFile)
	{
		_tprintf(_T("ERROR: Could not open \"%s\" (%lu)\n"), szFilename, GetLastError());
	}
	else
	{
		DWORD dwFileSize = GetFileSize(hFile, NULL);

		if (INVALID_FILE_SIZE == dwFileSize)
		{
			_tprintf(_T("ERROR: Could not get file size (%lu)\n"), GetLastError());
		}
		else
		{
			SIZE_T bufferSize = dwFileSize + 1 + strClipboardText.length() + (bSingleLine ? 0 : 2);

			HGLOBAL hGlobalClipMem = GlobalAlloc(GMEM_MOVEABLE, bufferSize); // +1 for null.

			if (NULL == hGlobalClipMem)
			{
				_tprintf(_T("ERROR: Could not allocate text memory (%lu)\n"), GetLastError());
			}
			else
			{
				char *pGlobalClipBuffer = reinterpret_cast<char *>( ::GlobalLock(hGlobalClipMem) );

				if (NULL == pGlobalClipBuffer)
				{
					_tprintf(_T("ERROR: Could not lock text memory (%lu)\n"), GetLastError());
				}
				else
				{
					_tcsncpy_s(pGlobalClipBuffer, bufferSize, strClipboardText.c_str(), _TRUNCATE);

					if (!bSingleLine)
					{
						_tcsncpy_s(pGlobalClipBuffer + strClipboardText.length(), bufferSize - strClipboardText.length(), "\r\n", _TRUNCATE);
					}

					DWORD dwReadBytes = 0;

					if (!ReadFile(hFile, pGlobalClipBuffer + strClipboardText.length() + (bSingleLine ? 0 : 2), dwFileSize, &dwReadBytes, NULL)
					||	dwReadBytes != dwFileSize)
					{
						_tprintf(_T("ERROR: Could not read the file (%lu)\n"), GetLastError());
					}
					else
					{
						reinterpret_cast<char *>(pGlobalClipBuffer)[bufferSize] = '\0'; // We allocated an extra byte for this above.

						GlobalUnlock(hGlobalClipMem);
						pGlobalClipBuffer = NULL;

						if (!OpenClipboard(NULL))
						{
							_tprintf(_T("ERROR: Could not open clipboard (%lu)\n"), GetLastError());
						}
						else
						{
							if (!EmptyClipboard())
							{
								_tprintf(_T("ERROR: Could not empty clipboard (%lu)\n"), GetLastError());
							}
							else
							{
								if (NULL == (SetClipboardData(CF_TEXT, hGlobalClipMem)))
								{
									_tprintf(_T("ERROR: Could not set clipboard (%lu)\n"), GetLastError());
								}
								else
								{
									hGlobalClipMem = NULL;
								}
							}

							if (!CloseClipboard())
							{
								_tprintf(_T("ERROR: Could not close clipboard (%lu)\n"), GetLastError());
							}
						}
					}

					if (NULL != pGlobalClipBuffer)
					{
						GlobalUnlock(hGlobalClipMem);
						pGlobalClipBuffer = NULL;
					}
				}

				if (NULL != hGlobalClipMem)
				{
					GlobalFree(hGlobalClipMem);
					hGlobalClipMem = NULL;
				}
			}

		}

		CloseHandle(hFile);
	}
}

void EmptyClip()
{
	if (!OpenClipboard(NULL))
	{
		_tprintf(_T("ERROR: Could not open clipboard (%lu)\n"), GetLastError());
	}
	else
	{
		if (!EmptyClipboard())
		{
			_tprintf(_T("ERROR: Could not empty clipboard (%lu)\n"), GetLastError());
		}

		if (!CloseClipboard())
		{
			_tprintf(_T("ERROR: Could not close clipboard (%lu)\n"), GetLastError());
		}
	}
}

void SetClipFromArguments(const int startArg, const int argc, _TCHAR* argv[], const bool bSingleLine, const bool bAppend)
{
	std::basic_string<TCHAR> strClipboardText;

	if (bAppend)
	{
		if (!GetClipboardText(strClipboardText))
		{
			strClipboardText.clear();
		}
	}

	int i;

	SIZE_T dwAllocSize = 1; // 1 for null.

	const bool bActuallyAppending = (!strClipboardText.empty());

	if (bActuallyAppending)
	{
		dwAllocSize += strClipboardText.length();

		if (startArg < argc)
		{
			if (bSingleLine)
			{
				dwAllocSize += 1; // +1 for space
			}
			else
			{
				dwAllocSize += 2; // +2 for \r\n
			}
		}
	}

	for (i = startArg; i < argc; i++)
	{
		dwAllocSize += strlen(argv[i]);
		
		if (i != argc-1)
		{
			if (bSingleLine)
			{
				dwAllocSize += 1; // +1 for space
			}
			else
			{
				dwAllocSize += 2; // +2 for \r\n
			}
		}
	}

	HGLOBAL hGlobalClipMem = GlobalAlloc(GMEM_MOVEABLE, dwAllocSize * sizeof(TCHAR));

	if (NULL == hGlobalClipMem)
	{
		_tprintf(_T("ERROR: Could not allocate text memory (%lu)\n"), GetLastError());
	}
	else
	{
		char *szGlobalClipBuffer = reinterpret_cast<char *>(GlobalLock(hGlobalClipMem));
		SIZE_T charsLeft = dwAllocSize;

		if (NULL == szGlobalClipBuffer)
		{
			_tprintf(_T("ERROR: Could not lock text memory (%lu)\n"), GetLastError());
		}
		else
		{
			if (bActuallyAppending)
			{
				_tcsncpy_s(szGlobalClipBuffer, charsLeft, strClipboardText.c_str(), _TRUNCATE);
				szGlobalClipBuffer += strClipboardText.length();
				charsLeft -= strClipboardText.length();

				if (startArg < argc)
				{
					if (bSingleLine)
					{
						_tcsncpy_s(szGlobalClipBuffer, charsLeft, _T(" "), _TRUNCATE);
						szGlobalClipBuffer += 1;
						charsLeft -= 1;
					}
					else
					{
						_tcsncpy_s(szGlobalClipBuffer, charsLeft, _T("\r\n"), _TRUNCATE);
						szGlobalClipBuffer += 2;
						charsLeft -= 2;
					}
				}
			}

			for (i = startArg; i < argc; i++)
			{
				_tcsncpy_s(szGlobalClipBuffer, charsLeft, argv[i], _TRUNCATE);
				SIZE_T strLen = _tcslen(argv[i]);
				szGlobalClipBuffer += strLen;
				charsLeft -= strLen;

				if (i != argc-1)
				{
					if (bSingleLine)
					{
						_tcsncpy_s(szGlobalClipBuffer, charsLeft, _T(" "), _TRUNCATE);
						szGlobalClipBuffer += 1;
						charsLeft -= 1;
					}
					else
					{
						_tcsncpy_s(szGlobalClipBuffer, charsLeft, _T("\r\n"), _TRUNCATE);
						szGlobalClipBuffer += 2;
						charsLeft -= 2;
					}
				}
			}

			*szGlobalClipBuffer = _T('\0'); // Only really needed when nothing is written, but harmless in other cases.

			GlobalUnlock(hGlobalClipMem);
			szGlobalClipBuffer = NULL;

			if (!OpenClipboard(NULL))
			{
				_tprintf(_T("ERROR: Could not open clipboard (%lu)\n"), GetLastError());
			}
			else
			{
				if (!EmptyClipboard())
				{
					_tprintf(_T("ERROR: Could not empty clipboard (%lu)\n"), GetLastError());
				}
				else
				{
#ifdef UNICODE
					if (NULL == (SetClipboardData(CF_UNICODETEXT, hGlobalClipMem)))
#else
					if (NULL == (SetClipboardData(CF_TEXT, hGlobalClipMem)))
#endif
					{
						_tprintf(_T("ERROR: Could not set clipboard (%lu)\n"), GetLastError());
					}
					else
					{
						hGlobalClipMem = NULL;
					}
				}

				if (!CloseClipboard())
				{
					_tprintf(_T("ERROR: Could not close clipboard (%lu)\n"), GetLastError());
				}
			}
		}

		if (NULL != hGlobalClipMem)
		{
			GlobalFree(hGlobalClipMem);
			hGlobalClipMem = NULL;
		}
	}
}

bool GetClipboardText(std::basic_string<TCHAR> &strClipboardText)
{
	bool bResult = false;
	strClipboardText.clear();

	if (!OpenClipboard(NULL))
	{
		_tprintf(_T("ERROR: Could not open clipboard (%lu)\n"), GetLastError());
	}
	else
	{
#ifdef UNICODE
		if (!::IsClipboardFormatAvailable(CF_UNICODETEXT))
#else
		if (!::IsClipboardFormatAvailable(CF_TEXT))
#endif
		{
			_tprintf(_T("WARNING: Cannot append as no text data in the clipboard.\n"));
		}
		else
		{
#ifdef UNICODE
			HANDLE hClipData = ::GetClipboardData(CF_UNICODETEXT);
#else
			HANDLE hClipData = ::GetClipboardData(CF_TEXT);
#endif
			if (NULL == hClipData)
			{
				_tprintf(_T("WARNING: Cannot append as unable to read from the clipboard (%lu)\n"), GetLastError());
			}
			else
			{
				const TCHAR *szBuffer = reinterpret_cast<const TCHAR *>( ::GlobalLock(hClipData) );

				if (NULL == szBuffer)
				{
					_tprintf(_T("WARNING: Cannot append as unable to lock the clipboard data (%lu)\n"), GetLastError());
				}
				else
				{
					strClipboardText = szBuffer;
					bResult = true;

					if (!::GlobalUnlock(hClipData))
					{
						_tprintf(_T("ERROR: Could not unlock the clipboard data (%lu)\n"), GetLastError());
					}
				}
			}
		}

		if (!CloseClipboard())
		{
			_tprintf(_T("ERROR: Could not close clipboard (%lu)\n"), GetLastError());
		}
	}

	return bResult;
}
